
using System;
using UnityEngine.Extensions;

namespace UnityEngine.UI.Windows.Components {
    
    using static HTMLAttributes;

    public interface IHTMLNodeRecevier {

        void SetInfo(HtmlAgilityPack.HtmlNode node);

    }
    
    public class HTMLPageComponent : WindowComponent {

        [System.Serializable]
        public struct Node {

            public string tag;
            public string @class;
            public GameObject source;

        }

        public Node[] nodes;
        public Node defaultNode;
        public HTMLPageTags[] tags;
        public TextAsset pageSource;
        public TextAsset[] cssSources;
        public SpriteMask particlesMask;
        public int minOrderDelta = -30;
        public int maxOrderDelta = 30;

        private GameObject root;
        private bool spriteMask;
        
        public void LateUpdate() {

            if (this.spriteMask == false) return;
            
            if (this.GetComponentState() != UnityEngine.UI.Windows.WindowObjectState.Showing &&
                this.GetComponentState() != UnityEngine.UI.Windows.WindowObjectState.Shown) {
                return;
            }

            if (this.GetWindow().GetActiveState() == UnityEngine.UI.Windows.ActiveState.Inactive) {
                return;
            }

            this.UpdateMask();

        }
        
        public override void OnWindowLayoutComplete() {

            base.OnWindowLayoutComplete();

            this.UpdateMask();

        }
        
        private void UpdateMask() {

            if (this.spriteMask == false) return;
            
            var canvas = this.GetWindow().GetCanvas();
            if (canvas == null || this.particlesMask == null) {
                return;
            }

            var maskTransform = (RectTransform)this.particlesMask.transform;
            maskTransform.SetParent(this.transform.parent);
            var ppu = this.particlesMask.sprite.pixelsPerUnit;
            maskTransform.localScale = new Vector3(maskTransform.rect.width * ppu, maskTransform.rect.height * ppu, 1.0f);
            this.particlesMask.backSortingLayerID = canvas.sortingLayerID;
            this.particlesMask.frontSortingLayerID = canvas.sortingLayerID;
            this.particlesMask.backSortingOrder = this.minOrderDelta + canvas.sortingOrder;
            this.particlesMask.frontSortingOrder = this.maxOrderDelta + canvas.sortingOrder;
            maskTransform.anchorMin = new Vector2(0f, 0f);
            maskTransform.anchorMax = new Vector2(1f, 1f);
            maskTransform.localPosition = Vector3.zero;
            maskTransform.sizeDelta = Vector2.zero;

        }

        public override void OnShowBegin() {
            
            base.OnShowBegin();
            
            this.Rebuild();
            this.UpdateMask();

        }

        public void SetInfo(string data) {
            
            var htmlDoc = new HtmlAgilityPack.HtmlDocument();
            htmlDoc.LoadHtml(data);
            for (int i = this.subComponents.Count - 1; i >= 0; --i) {
                var obj = this.subComponents[i];
                this.UnregisterSubComponent(obj);
                obj.Recycle();
            }

            if (this.cssSources != null) {
                foreach (var cssSource in this.cssSources) {
                    var css = new ExCSS.StylesheetParser(includeUnknownRules: true, includeUnknownDeclarations: true, tolerateInvalidValues: true, tolerateInvalidConstraints: true, tolerateInvalidSelectors: true);
                    var cssObj = css.Parse(cssSource.text);
                    foreach (var obj in cssObj.StyleRules) {
                        var nodes = htmlDoc.DocumentNode.QuerySelectorAll(obj.SelectorText);
                        foreach (var node in nodes) {
                            foreach (var style in obj.Style) {
                                node.SetAttributeValue(style.Name, style.Value);
                            }
                        }
                    }
                }
            }

            if (this.root != null) {
                GameObject.DestroyImmediate(this.root);
            }

            var rootNode = this.CreateNode(htmlDoc.DocumentNode, null, this.transform);
            this.spriteMask = HTMLUtils.GetBool(htmlDoc.DocumentNode, sprite_mask, false);
            this.particlesMask.gameObject.SetActive(this.spriteMask);
            this.root = rootNode;
            this.CreateNodes(htmlDoc.DocumentNode, rootNode.transform);
            
        }
        
        [ContextMenu("Rebuild")]
        public void Rebuild() {

            if (this.pageSource == null) return;
            var data = this.pageSource.text;
            this.SetInfo(data);
            
        }

        private void CreateNodes(HtmlAgilityPack.HtmlNode root, Transform trRoot) {
            
            foreach (var node in root.ChildNodes) {
                if (CheckNode(node) == false) continue;
                var go = this.CreateNode(node, root, trRoot);
                if (go == null) continue;
                this.CreateNodes(node, go.transform);
            }
            
        }

        private GameObject CreateNode(HtmlAgilityPack.HtmlNode node, HtmlAgilityPack.HtmlNode parentNode, Transform transformParent) {

            {
                // Check if node is a text node and it has no children
                if (node.NodeType == HtmlAgilityPack.HtmlNodeType.Text &&
                    node.HasChildNodes == false) {
                    if (transformParent.GetComponent<ITextComponent>() != null /*&&
                        string.IsNullOrEmpty(node.Attributes["background-color"]?.Value) == true &&
                        string.IsNullOrEmpty(node.Attributes["background-alpha"]?.Value) == false*/) {
                        // set to parent node
                        var text = GetOrAdd<ITextComponent, TextComponent>(transformParent.gameObject);
                        SetTextWithLocalization(text, parentNode);
                        // clean up node
                        return null;
                    }
                }
            }

            var parentObj = transformParent.GetComponent<WindowComponent>();
            if (parentObj == null) parentObj = transformParent.GetComponentInParent<WindowComponent>();
            
            var source = this.FindSource(node);
            var obj = GameObject.Instantiate(source, transformParent);
            var nodeName = node.Name;
            var customName = node.Attributes[node_name]?.Value;
            if (string.IsNullOrEmpty(customName) == false) {
                nodeName = customName;
            }
            obj.name = $"<{nodeName}>";
            {
                if (node.NodeType == HtmlAgilityPack.HtmlNodeType.Text &&
                    node.HasChildNodes == false) {
                    var text = GetOrAdd<TextComponent>(obj);
                    SetTextWithLocalization(text, node);
                }
            }
            var hasChildNodes = node.HasChildNodes;
            if (hasChildNodes == true) {
                var childNodes = node.ChildNodes;
                if (childNodes.Count == 1) {
                    if (childNodes[0].NodeType == HtmlAgilityPack.HtmlNodeType.Text &&
                        childNodes[0].HasChildNodes == false &&
                        obj.GetComponent<ITextComponent>() != null) {
                        hasChildNodes = false;
                    }
                }
            }
            
            ApplyLayout(obj, node, hasChildNodes);
            ApplyInteractable(obj, node);
            ApplyBorders(obj, node);
            ApplyColors(obj, node);
            ApplyText(obj, node);
            
            if (obj.TryGetComponent(out WindowComponent comp) == true) {
                parentObj.RegisterSubComponent(comp);
                comp.Setup(parentObj.GetWindow());
            }

            if (Application.isPlaying == true) {
                if (obj.TryGetComponent(out IHTMLNodeRecevier receiver) == true) {
                    receiver.SetInfo(node);
                }
            }
            
            return obj;

        }

        private static void ApplyInteractable(GameObject obj, HtmlAgilityPack.HtmlNode node) {
            var intr = HTMLUtils.GetBool(node, interactable, true);
            if (intr == false) {
                var canvasGroup = GetOrAdd<CanvasGroup>(obj);
                canvasGroup.interactable = false;
                canvasGroup.blocksRaycasts = false;
            }
        }

        private static void ApplyLayout(GameObject obj, HtmlAgilityPack.HtmlNode node, bool hasChildNodes) {

            LayoutGroup group = null;
            if (hasChildNodes == true) {
                group = GetLayoutGroup(node, obj);
                ApplyLayoutGroup(group, node);
            }

            var rectPadding = group != null ? new Vector4(group.padding.left, group.padding.top, group.padding.right, group.padding.bottom) : Vector4.zero;
            {
                if (TryGetAttributeIntValue(node, padding, out var paddingValue) == true) {
                    rectPadding.x = rectPadding.y = rectPadding.z = rectPadding.w = paddingValue;
                }
                if (TryGetAttributeIntValue(node, padding_left, out var paddingLeftValue) == true) {
                    rectPadding.x = paddingLeftValue;
                }
                if (TryGetAttributeIntValue(node, padding_right, out var paddingRightValue) == true) {
                    rectPadding.z = paddingRightValue;
                }
                if (TryGetAttributeIntValue(node, padding_bottom, out var paddingBottomValue) == true) {
                    rectPadding.w = paddingBottomValue;
                }
                if (TryGetAttributeIntValue(node, padding_top, out var paddingTopValue) == true) {
                    rectPadding.y = paddingTopValue;
                }
                
                if (hasChildNodes == true) {
                    group.childAlignment = HTMLUtils.GetEnum(node, child_alignment, group.childAlignment);
                    group.padding.left = (int)rectPadding.x;
                    group.padding.bottom = (int)rectPadding.w;
                    group.padding.right = (int)rectPadding.z;
                    group.padding.top = (int)rectPadding.y;
                }
            }
            {
                var layoutElement = GetOrAdd<LayoutElement>(obj);
                layoutElement.ignoreLayout = HTMLUtils.GetString(node, position, position_relative) == position_absolute;
                layoutElement.flexibleWidth = HTMLUtils.GetFloat(node, flex_width, layoutElement.flexibleWidth);
                layoutElement.flexibleHeight = HTMLUtils.GetFloat(node, flex_height, layoutElement.flexibleHeight);
                layoutElement.preferredWidth = HTMLUtils.GetFloat(node, width, layoutElement.preferredWidth);
                layoutElement.preferredHeight = HTMLUtils.GetFloat(node, height, layoutElement.preferredHeight);
                layoutElement.minWidth = HTMLUtils.GetFloat(node, min_width, layoutElement.minWidth);
                layoutElement.minHeight = HTMLUtils.GetFloat(node, min_height, layoutElement.minHeight);
                if (layoutElement.ignoreLayout == true) {
                    var rect = (RectTransform)layoutElement.transform;
                    rect.offsetMin = new Vector2(rectPadding.x, rectPadding.w);
                    rect.offsetMax = new Vector2(-rectPadding.z, -rectPadding.y);
                }
            }
            
        }

        private struct BorderParams {

            public string name;
            public string color;
            public string alpha;
            public string size;
            public BorderParams(string name, string color, string alpha, string size) {
                this.name = name;
                this.color = color;
                this.alpha = alpha;
                this.size = size;
            }

        }

        private static void ApplyBorders(GameObject obj, HtmlAgilityPack.HtmlNode node) {

            static void LeftBorderSetup(Image x) {
                x.rectTransform.anchorMin = new Vector2(0f, 0f);
                x.rectTransform.anchorMax = new Vector2(0f, 1f);
            }

            static void RightBorderSetup(Image x) {
                x.rectTransform.anchorMin = new Vector2(1f, 0f);
                x.rectTransform.anchorMax = new Vector2(1f, 1f);
            }

            static void TopBorderSetup(Image x) {
                x.rectTransform.anchorMin = new Vector2(0f, 1f);
                x.rectTransform.anchorMax = new Vector2(1f, 1f);
            }

            static void BottomBorderSetup(Image x) {
                x.rectTransform.anchorMin = new Vector2(0f, 0f);
                x.rectTransform.anchorMax = new Vector2(1f, 0f);
            }

            static void ApplySizeLeft(Image x, float size) {
                x.rectTransform.sizeDelta = new Vector2(size, x.rectTransform.sizeDelta.y);
                x.rectTransform.pivot = new Vector2(0f, 0.5f);
            }

            static void ApplySizeRight(Image x, float size) {
                x.rectTransform.sizeDelta = new Vector2(size, x.rectTransform.sizeDelta.y);
                x.rectTransform.pivot = new Vector2(1f, 0.5f);
            }

            static void ApplySizeTop(Image x, float size) {
                x.rectTransform.sizeDelta = new Vector2(-size * 2f, size);
                x.rectTransform.pivot = new Vector2(0.5f, 1f);
            }

            static void ApplySizeBottom(Image x, float size) {
                x.rectTransform.sizeDelta = new Vector2(-size * 2f, size);
                x.rectTransform.pivot = new Vector2(0.5f, 0f);
            }

            static Image DefaultSetup(Image border, System.Action<Image> setupAction, System.Action<Image, float> setupSizeAction, BorderParams p, GameObject obj, HtmlAgilityPack.HtmlNode node) {
                
                if (HTMLUtils.GetColor(node, p.color, out var color) == true) {
                    border = border ? border : AddChild(obj, p.name, setupAction);
                    border.color = color;
                }

                if (HTMLUtils.GetFloat(node, p.size, out var size) == true) {
                    border = border ? border : AddChild(obj, p.name, setupAction);
                    setupSizeAction.Invoke(border, size);
                }

                if (HTMLUtils.GetFloat(node, p.alpha, out var alpha) == true) {
                    border = border ? border : AddChild(obj, p.name, setupAction);
                    var c = border.color;
                    c.a = alpha;
                    border.color = c;
                }

                return border;
            }

            Image borderLeft = null;
            Image borderRight = null;
            Image borderTop = null;
            Image borderBottom = null;
            {
                if (HTMLUtils.Has(node, border_size) == true ||
                    HTMLUtils.Has(node, border_color) == true ||
                    HTMLUtils.Has(node, border_alpha) == true) {

                    borderLeft = DefaultSetup(null, LeftBorderSetup, ApplySizeLeft, new BorderParams(border_left, border_color, border_alpha, border_size), obj, node);
                    borderRight = DefaultSetup(null, RightBorderSetup, ApplySizeRight, new BorderParams(border_right, border_color, border_alpha, border_size), obj, node);
                    borderTop = DefaultSetup(null, TopBorderSetup, ApplySizeTop, new BorderParams(border_top, border_color, border_alpha, border_size), obj, node);
                    borderBottom = DefaultSetup(null, BottomBorderSetup, ApplySizeBottom, new BorderParams(border_bottom, border_color, border_alpha, border_size), obj, node);

                }
            }

            DefaultSetup(borderLeft, LeftBorderSetup, ApplySizeLeft, new BorderParams(border_left, border_left_color, border_left_alpha, border_left_size), obj, node);
            DefaultSetup(borderRight, RightBorderSetup, ApplySizeRight, new BorderParams(border_right, border_right_color, border_right_alpha, border_right_size), obj, node);
            DefaultSetup(borderTop, TopBorderSetup, ApplySizeTop, new BorderParams(border_top, border_top_color, border_top_alpha, border_top_size), obj, node);
            DefaultSetup(borderBottom, BottomBorderSetup, ApplySizeBottom, new BorderParams(border_bottom, border_bottom_color, border_bottom_alpha, border_bottom_size), obj, node);

        }

        private static T AddChild<T>(GameObject root, string name, System.Action<T> setupAction = null) where T : Component {
            
            var graphic = new GameObject(name, typeof(RectTransform), typeof(LayoutElement), typeof(T)).GetComponent<T>();
            graphic.transform.SetParent(root.transform);
            var rect = graphic.GetComponent<RectTransform>();
            rect.pivot = new Vector2(0.5f, 0.5f);
            rect.anchorMin = new Vector2(0f, 0f);
            rect.anchorMax = new Vector2(1f, 1f);
            rect.localPosition = Vector3.zero;
            rect.sizeDelta = Vector2.zero;
            rect.localScale = Vector3.one;
            setupAction?.Invoke(graphic);
            var layoutElement = graphic.GetComponent<LayoutElement>();
            layoutElement.ignoreLayout = true;
            return graphic;

        }

        private static void ApplyColors(GameObject obj, HtmlAgilityPack.HtmlNode node) {

            var group = obj.GetComponent<CanvasGroup>();
            var alpha = HTMLUtils.GetFloat(node, opacity, 1f);
            if (alpha != 1f || group != null) {
                if (group == null) group = obj.AddComponent<CanvasGroup>();
                group.alpha = alpha;
            }

            if (node.Attributes.Contains(background_color) == true ||
                node.Attributes.Contains(background_alpha) == true) {
                var graphic = obj.GetComponent<Graphic>();
                if (graphic == null) graphic = obj.AddComponent<Image>();
                /*if (graphic == null) {
                    graphic = new GameObject("Background", typeof(RectTransform), typeof(Image)).GetComponent<Image>();
                    graphic.transform.SetParent(obj.transform);
                    graphic.rectTransform.anchorMin = new Vector2(0f, 0f);
                    graphic.rectTransform.anchorMax = new Vector2(1f, 1f);
                    graphic.rectTransform.localPosition = Vector3.zero;
                    graphic.rectTransform.sizeDelta = Vector2.zero;
                    graphic.rectTransform.localScale = Vector3.one;
                }*/

                var c = graphic.color;
                c = HTMLUtils.GetColor(node, background_color, c);
                c.a = HTMLUtils.GetFloat(node, background_alpha, graphic.color.a);
                graphic.color = c;
            }
            
        }

        private static void ApplyText(GameObject obj, HtmlAgilityPack.HtmlNode node) {

            if (node.Attributes.Contains(text_color) == true) {
                var graphic = obj.GetComponent<TextComponent>();
                if (graphic != null) {
                    graphic.SetTextColor(HTMLUtils.GetColor(node, text_color, graphic.GetTextColor()));
                }
            }
            if (node.Attributes.Contains(text_alpha) == true) {
                var graphic = obj.GetComponent<TextComponent>();
                if (graphic != null) {
                    graphic.SetTextAlpha(HTMLUtils.GetFloat(node, text_alpha, graphic.GetTextAlpha()));
                }
            }

        }

        private static void ApplyLayoutGroup(LayoutGroup group, HtmlAgilityPack.HtmlNode node) {
            if (group is HorizontalOrVerticalLayoutGroup horizontalOrVerticalLayoutGroup) {
                horizontalOrVerticalLayoutGroup.childControlWidth = HTMLUtils.GetBool(node, child_control_width, true);
                horizontalOrVerticalLayoutGroup.childControlHeight = HTMLUtils.GetBool(node, child_control_height, true);
                horizontalOrVerticalLayoutGroup.childScaleWidth = HTMLUtils.GetBool(node, child_scale_width, false);
                horizontalOrVerticalLayoutGroup.childScaleHeight = HTMLUtils.GetBool(node, child_scale_height, false);
                horizontalOrVerticalLayoutGroup.childForceExpandWidth = HTMLUtils.GetBool(node, child_expand_width, false);
                horizontalOrVerticalLayoutGroup.childForceExpandHeight = HTMLUtils.GetBool(node, child_expand_height, false);
                horizontalOrVerticalLayoutGroup.spacing = HTMLUtils.GetFloat(node, child_spacing, horizontalOrVerticalLayoutGroup.spacing);
                horizontalOrVerticalLayoutGroup.reverseArrangement = HTMLUtils.GetBool(node, child_reverse, horizontalOrVerticalLayoutGroup.reverseArrangement);
            } else if (group is UnityEngine.UI.Windows.Extensions.FlowLayoutGroup flowLayoutGroup) {
                flowLayoutGroup.Spacing = HTMLUtils.GetFloat(node, child_spacing, flowLayoutGroup.Spacing);
                flowLayoutGroup.ChildForceExpandWidth = HTMLUtils.GetBool(node, child_expand_width, false);
                flowLayoutGroup.ChildForceExpandHeight = HTMLUtils.GetBool(node, child_expand_height, false);
            } else if (group is GridLayoutGroup gridLayoutGroup) {
                var spacing = gridLayoutGroup.spacing;
                spacing.y = HTMLUtils.GetFloat(node, child_row_spacing, spacing.y);
                spacing.x = HTMLUtils.GetFloat(node, child_column_spacing, spacing.x);
                gridLayoutGroup.spacing = spacing;
                gridLayoutGroup.startCorner = HTMLUtils.GetEnum(node, start_corner, gridLayoutGroup.startCorner);
                gridLayoutGroup.startAxis = HTMLUtils.GetEnum(node, start_axis, gridLayoutGroup.startAxis);
                gridLayoutGroup.childAlignment = HTMLUtils.GetEnum(node, child_alignment, gridLayoutGroup.childAlignment);
                gridLayoutGroup.constraintCount = HTMLUtils.GetInt(node, constraint_count, gridLayoutGroup.constraintCount);
                gridLayoutGroup.constraint = HTMLUtils.GetEnum(node, constraint, gridLayoutGroup.constraint);
            } else if (group is UnityEngine.UI.Extensions.TableLayoutGroup tableLayoutGroup) {
                tableLayoutGroup.StartCorner = HTMLUtils.GetEnum(node, start_corner, tableLayoutGroup.StartCorner);
                tableLayoutGroup.ColumnSpacing = HTMLUtils.GetFloat(node, child_spacing, tableLayoutGroup.ColumnSpacing);
                tableLayoutGroup.RowSpacing = HTMLUtils.GetFloat(node, child_spacing, tableLayoutGroup.RowSpacing);
                tableLayoutGroup.ColumnSpacing = HTMLUtils.GetFloat(node, child_column_spacing, tableLayoutGroup.ColumnSpacing);
                tableLayoutGroup.RowSpacing = HTMLUtils.GetFloat(node, child_row_spacing, tableLayoutGroup.RowSpacing);
                tableLayoutGroup.MinimumRowHeight = HTMLUtils.GetFloat(node, min_row_height, tableLayoutGroup.MinimumRowHeight);
                tableLayoutGroup.FlexibleRowHeight = HTMLUtils.GetBool(node, flex_row_height, tableLayoutGroup.FlexibleRowHeight);
                tableLayoutGroup.ColumnWidths = HTMLUtils.GetFloatArray(node, column_widths, tableLayoutGroup.ColumnWidths);
            }
        }

        private static LayoutGroup GetLayoutGroup(HtmlAgilityPack.HtmlNode node, GameObject obj) {
            var layoutGroup = obj.GetComponent<LayoutGroup>();
            if (layoutGroup != null) return layoutGroup;
            if (node.Attributes[layout]?.Value == layout_flow) {
                var layout = GetOrAdd<LayoutElement>(obj);
                layout.flexibleWidth = 1f;
                return GetOrAdd<UnityEngine.UI.Windows.Extensions.FlowLayoutGroup>(obj);
            }
            if (node.Attributes[layout]?.Value == layout_horizontal) {
                return GetOrAdd<HorizontalLayoutGroup>(obj);
            }
            if (node.Attributes[layout]?.Value == layout_table) {
                return GetOrAdd<UnityEngine.UI.Extensions.TableLayoutGroup>(obj);
            }
            if (node.Attributes[layout]?.Value == layout_grid) {
                return GetOrAdd<GridLayoutGroup>(obj);
            }
            return GetOrAdd<VerticalLayoutGroup>(obj);
        }

        private static void SetTextWithLocalization(ITextComponent text, HtmlAgilityPack.HtmlNode node) {
            var nodeAttr = node;
            if (nodeAttr.NodeType == HtmlAgilityPack.HtmlNodeType.Text) {
                nodeAttr = nodeAttr.ParentNode;
            }
            
            var str = HTMLUtils.GetString(nodeAttr, localization_key);
            if (str == null) {
                if (long.TryParse(node.InnerText, out var longValue) == true) {
                    text.SetValue(longValue);
                } else {
                    text.SetText(node.InnerText);
                }
            } else {
                var key = new UnityEngine.UI.Windows.Plugins.Localization.LocalizationKey(str);
                text.SetText(key);
            }

            text.SetTextAlignment(HTMLUtils.GetEnum(nodeAttr, text_alignment, text.GetTextAlignment()));
            text.SetFontSize(HTMLUtils.GetInt(nodeAttr, text_size, text.GetFontSize()));
            text.SetLineSpacing(HTMLUtils.GetFloat(nodeAttr, text_line_spacing, text.GetLineSpacing()));
            if (HTMLUtils.GetEnum(nodeAttr, text_value_format, out TextValueFormat textValueFormat) == true) {
                text.SetValueFormat(textValueFormat);
            }
            text.SetBestFitState(HTMLUtils.GetBool(nodeAttr, text_bestfit, text.GetBestFitState()));
            text.SetBestFitMinSize(HTMLUtils.GetInt(nodeAttr, text_bestfit_min_size, text.GetBestFitMinSize()));
            text.SetBestFitMaxSize(HTMLUtils.GetInt(nodeAttr, text_bestfit_max_size, text.GetBestFitMaxSize()));
            if (HTMLUtils.GetEnum(nodeAttr, text_fulltext_format, out FullTextFormat fullTextFormat) == true) {
                text.SetFullTextFormat(fullTextFormat);
            }
            if (HTMLUtils.GetEnum(nodeAttr, text_vertical_wrap_mode, out VerticalWrapMode verticalWrapMode) == true) {
                text.SetTextVerticalOverflow(verticalWrapMode);
            }
            if (HTMLUtils.GetEnum(nodeAttr, text_horizontal_wrap_mode, out HorizontalWrapMode horizontalWrapMode) == true) {
                text.SetTextHorizontalOverflow(horizontalWrapMode);
            }
        }

        private static bool CheckNode(HtmlAgilityPack.HtmlNode node) {
            if (node.NodeType == HtmlAgilityPack.HtmlNodeType.Comment) return false;
            if (node.NodeType == HtmlAgilityPack.HtmlNodeType.Text &&
                string.IsNullOrWhiteSpace(node.InnerText) == true &&
                string.IsNullOrWhiteSpace(node.Attributes[localization_key]?.Value) == true) {
                return false;
            }
            return true;
        }

        private static T GetOrAdd<T>(GameObject obj) where T : Component {
            var group = obj.GetComponent<T>();
            if (group == null) group = obj.AddComponent<T>();
            return group;
        }

        private static TLook GetOrAdd<TLook, TCreate>(GameObject obj) where TLook : IComponent where TCreate : Component, TLook {
            var group = obj.GetComponent<TLook>();
            if (group == null) group = obj.AddComponent<TCreate>();
            return group;
        }

        private static bool TryGetAttributeIntValue(HtmlAgilityPack.HtmlNode node, string attr, out int value) {
            var padding = node.Attributes[attr]?.Value;
            if (int.TryParse(padding, out value) == true) {
                return true;
            }
            return false;
        }

        private GameObject FindSource(HtmlAgilityPack.HtmlNode node) {

            foreach (var item in this.nodes) {
                if (item.tag == node.Name &&
                    string.IsNullOrEmpty(item.@class) == false &&
                    node.HasClass(item.@class) == true) {
                    return item.source;
                }
            }
            
            foreach (var item in this.nodes) {
                if (item.tag == node.Name &&
                    string.IsNullOrEmpty(item.@class) == true) {
                    return item.source;
                }
            }

            foreach (var item in this.tags) {
                if (item == null) continue;
                var go = item.FindSource(node);
                if (go != null) return go;
            }

            return this.defaultNode.source;

        }

    }

}